(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)
(* ignore the following; they are all for internal debugging purposes: *)
type constant = Num.num
type vartype =
    RealNode of Dag.node
  | DebugNode of string
type table = Num.num array
type element =
    Constant of constant
  | TimesVar of constant * vartype
  | TimesMin of constant * (table option * expr) list
  | TimesMax of constant * (table option * expr) list
  | Infinity
  | NegInfinity
and expr = element list (* implicitly a sum of 1 or more things *)

val sup : expr array -> expr -> expr
val inf : expr array -> expr -> expr
val simp : expr -> expr
val flatten : expr -> expr
val const_collect_min_convert : expr -> expr
val check_regressions : unit -> unit
val eqns : expr list
val makeExprFromDag : Dag.node -> expr
val makeExprFromInput : Input.exp -> expr
val string_of_expr : expr -> string
val upper : expr array -> vartype -> expr
val lower : expr array -> vartype -> expr
val make_multiple_minmaxes_into_one : expr -> expr
val mult_minmaxes_by_const : expr -> expr

exception RegressionFailure of expr * string

(* ok, don't ignore the things after this: *)

val initialize : Satproc.ident -> Satproc.entries
val simp_and_subtract_is_zero : Input.exp -> Input.exp -> bool
val simp_and_subtract_is_at_least_zero : Input.exp -> Input.exp -> bool
val simplify : Input.exp -> Input.exp
