(*
 * Description:
 *
 * Copyright (c) 2000 by
 *  David Blei			blei@cs.berkeley.edu
 *  Chris Harrelson		chrishtr@cs.berkeley.edu
 *  Ranjit Jhala		jhala@cs.berkeley.edu
 *  Rupak Majumdar		rupak@cs.berkeley.edu
 *  George C. Necula	necula@cs.berkeley.edu
 *  Shree P. Rahul		sprahul@cs.berkeley.edu
 * 	Wes Weimer			weimer@cs.berkeley.edu
 *  Dror Weitz			dror@cs.berkeley.edu
 *   
 * All rights reserved.  Permission to use, copy, modify and distribute
 * this software for research purposes only is hereby granted, 
 * provided that the following conditions are met: 
 * 1.  Redistributions of source code must retain the above copyright notice, 
 * this list of conditions and the following disclaimer. 
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 * this list of conditions and the following disclaimer in the documentation 
 * and/or other materials provided with the distribution. 
 * 3. The name of the authors may not be used to endorse or promote products derived from 
 * this software without specific prior written permission. 
 *
 * DISCLAIMER:
 * THIS SOFTWARE IS PROVIDED BY THE AUTHORS ``AS IS'' AND ANY EXPRESS OR IMPLIED 
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF 
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE AUTHORS BE LIABLE FOR ANY DIRECT, INDIRECT, 
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS 
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY 
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 *
 *)

type info

val current      : info ref             (* The current error info *)
val startFile    : string -> unit       (* Record the start of a new file *)
val startNewline : int -> unit          (* Record a newline at a given 
                                         * position *)

val getLineCol : info -> int -> string
val fileName   : info -> string

val logChannel : out_channel ref
val proofChannel : out_channel ref

val debugFlag  : bool ref               (* If set then print debugging info *)

val theLexbuf     : Lexing.lexbuf ref

exception Error


   (* Error reporting. All of these functions take same arguments as a 
    * VampyrePretty.eprintf. They raise the exception Error after they print their 
    * stuff. However, their type indicates that they return a "VampyrePretty.doc" 
    * (due to the need to use the built-in type "format") return a doc. Thus 
    * use as follows:  E.s (E.bug "different lengths (%d != %d)" l1 l2)
     *)
val error         : ('a,unit,VampyrePretty.doc) format -> 'a
val bug           : ('a,unit,VampyrePretty.doc) format -> 'a
val unimp         : ('a,unit,VampyrePretty.doc) format -> 'a
val s             : VampyrePretty.doc -> 'a

   (* Like error but does not raise an Error. Use: ignore (E.warn ...) *)
val warn         : ('a,unit,VampyrePretty.doc) format -> 'a

   (* Print something to the logChannel *)
val log           : ('a,unit,VampyrePretty.doc) format -> 'a

   (* All of the error and warning reporting functions can also print a 
    * context. To register a context printing function use "pushContext". To 
    * remove the last registered one use "popContext". If one of the error 
    * reporting functions is called it will invoke all currently registered 
    * context reporting functions in the reverse order they were registered. *)
val pushContext  : (unit -> VampyrePretty.doc) -> unit
val popContext   : unit -> unit




